import os
import numpy as np
from numpy.typing import NDArray
import json
import math
from scipy import fft
from typing import Any
from cratermaker.components.morphology import Morphology
from cratermaker.components.morphology.simplemoon import SimpleMoon

@Morphology.register("du2025")
class Du2025(SimpleMoon):
    """
    An operations class for computing the morphology of a crater and applying it to a surface mesh. This uses the morphology model of Du et al. 2025a,b

    Parameters
    ----------
    **kwargs : Any
        Additional keyword arguments to be passed to internal functions.
    """
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)


    def _get_1D_power_spectral_density(self,feature,psd_coef,num_psd_component_effec=100) -> NDArray:
        """
        This method construct a 1D power spectral density.
        Coeffcients are from [1]_.
       
        Parameters
        ----------
        feature : string
            For a 1D feature, choose from ejecta, rim, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 1D power spectral density
        num_psd_component_effec : int
            Only reconstrut the sine waves with wavelengths smaller than 2pi/num_psd_component_effec to improve computational efficiency

        References
        ----------
        .. [1] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2024). Spectral analysis of the morphology of fresh lunar craters I: Rim crest, floor, and rim flank outlines. Journal of Geophysical Research: Planets, 129(11), e2024JE008357. https://doi.org/10.1029/2024JE008357

        """
        # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        with open(os.path.join(os.pardir,"components","psd_coef.json")) as f:
            psd_coef = json.load(f)
        # ------------------------------------------------------------------------------------------------------------------
        num_psd_component=5000
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.final_diameter<psd_coef["1D"][feature]["slope_12"]["D_tie"]:
            slope_12 = psd_coef["1D"][feature]["slope_12"]["k1"] * self.crater.final_diameter + psd_coef["1D"][feature]["slope_12"]["b1"]
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_1"]
        else:
            slope_12 = psd_coef["1D"][feature]["slope_12"]["k2"] * self.crater.final_diameter + psd_coef["1D"][feature]["slope_12"]["b2"]
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["1D"][feature]["bp2_x"]["D_tie"]:
            bp2_x = psd_coef["1D"][feature]["bp2_x"]["k1"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp2_x"]["b1"]
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_1"]
        else:
            bp2_x = psd_coef["1D"][feature]["bp2_x"]["k2"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp2_x"]["b2"]
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["1D"][feature]["bp2_y"]["D_tie"]:
            bp2_y = psd_coef["1D"][feature]["bp2_y"]["k1"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp2_y"]["b1"]
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_1"]
        else:
            bp2_y = psd_coef["1D"][feature]["bp2_y"]["k2"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp2_y"]["b2"]
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["1D"][feature]["bp3_y"]["D_tie"]:
            bp3_y = psd_coef["1D"][feature]["bp3_y"]["k1"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp3_y"]["b1"]
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_1"]
        else:
            bp3_y = psd_coef["1D"][feature]["bp3_y"]["k2"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp3_y"]["b2"]
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["1D"][feature]["bp4_y"]["D_tie"]:
            bp4_y = psd_coef["1D"][feature]["bp4_y"]["k1"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp4_y"]["b1"]
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_1"]
        else:
            bp4_y = psd_coef["1D"][feature]["bp4_y"]["k2"] * self.crater.final_diameter + psd_coef["1D"][feature]["bp4_y"]["b2"]
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_2"]
        slope_12 = np.random.normal(slope_12, slope_12_sigma)
        bp2_x = np.random.normal(bp2_x, bp2_x_sigma)
        bp2_y = np.random.normal(bp2_y, bp2_y_sigma)
        bp3_y = np.random.normal(bp3_y, bp3_y_sigma)
        bp4_y = np.random.normal(bp4_y, bp4_y_sigma)
        psd_sigma=psd_coef["1D"][feature]["psd_sigma"]
        # ------------------------------------------------------------------------------------------------------------------
        bp4_x = math.log10(2*math.pi)
        bp3_x =  math.log10(10**bp4_x / 2)
        interval=2*math.pi/num_psd_component
        freq = fft.fftfreq(num_psd_component, interval)
        wavelength = 1 / freq[1:num_psd_component//2]
        psd = np.zeros((num_psd_component//2 - 1, 2))
        psd[:, 0] = wavelength
        # ----------------------------------------------------------------------------------------------------------------------
        bp2_x_index = int(  10**bp4_x/10**bp2_x-1)
        k_23=(bp3_y-bp2_y)/(bp3_x-bp2_x)
        b_23 = bp3_y-k_23*bp3_x
        k_12=slope_12
        b_12 = bp2_y-k_12*bp2_x
        #-----------------------------------------------------------------------------------------------------------------------
        psd[0,1] =10**bp4_x
        psd[1,1]=10**bp3_y
        psd[2:bp2_x_index+1,1] =10**(k_23*np.log10(psd[2:bp2_x_index+1, 0])+b_23)
        psd[bp2_x_index+1:, 1] =10**(k_12*np.log10(psd[bp2_x_index+1:, 0])+b_12)
        # ----------------------------------------------------------------------------------------------------------------------
        psd=psd[:num_psd_component_effec]
        psd_log = np.log10(psd[:, 1])
        psd_log += np.random.normal(0, psd_sigma, psd_log.shape) 
        psd[:, 1] = 10 ** psd_log
        psd=np.flipud(psd)
        return psd


    def _get_2D_power_spectral_density(self,feature,psd_coef,ejecta_radius_norm,floor_radius_norm,max_effec_freq) -> tuple[NDArray,NDArray,NDArray,NDArray,NDArray,NDArray]:
        """
        This method constructs a 2D power spectral density.
        Coeffcients are from [1]_.
       
        Parameters
        ----------
        feature : string
            For a 2D feature, choose from ejecta, wall, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 2D power spectral density
        ejecta_radius_norm : float
            The radius of the continuous ejecta normalized by the crater radius
        floor_radius_norm : float
            The radius of the floor normalized by the crater radius
        max_effec_freq : int
            Only reconstrut the sine waves with frequencies smaller than a maxmium effective frequency to improve computational efficiency 

        References
        ----------
        .. [1] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2025). Spectral Analysis of the Morphology of Fresh Lunar Craters II: Two-Dimensional Surface Elevations of the Continuous Ejecta, Wall, and Floor. Journal of Geophysical Research: Planets

        """
         # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        with open(os.path.join(os.pardir,"components","psd_coef.json")) as f:
            psd_coef = json.load(f) # read in from _init_?
        # ------------------------------------------------------------------------------------------------------------------
        if feature == "ejecta" : max_effec_freq = 40
        if feature == "wall" : max_effec_freq = 40
        if feature == "floor" : max_effec_freq = 100
        # ------------------------------------------------------------------------------------------------------------------
        if feature=="ejecta":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min=1
            r_max=ejecta_radius_norm
            area=(r_max - r_min) * (theta_max - theta_min)
            theta_number = 2000
            r_number = 400
        if feature == "wall":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min = floor_radius_norm
            r_max = 1
            area= (r_max - r_min) * (theta_max - theta_min)
            theta_number = 1000
            r_number = 200
        if feature == "floor":
            r_min = -floor_radius_norm/math.sqrt(2)
            r_max = floor_radius_norm/math.sqrt(2)
            theta_min = r_min
            theta_max = r_max
            area=(r_max - r_min) ** 2
            theta_number = 512
            r_number = 512
        # ------------------------------------------------------------------------------------------------------------------
        side_r=np.linspace(0, r_max-r_min, num=r_number, endpoint=False)
        side_theta=np.linspace(0, theta_max-theta_min, num=theta_number, endpoint=False)
        grid_theta, grid_r = np.meshgrid( side_theta,side_r)
        side_freq_theta = np.fft.fftfreq(theta_number, side_theta[1] - side_theta[0])
        side_freq_r = np.fft.fftfreq(r_number, side_r[1] - side_r[0])
        freq_theta, freq_r = np.meshgrid(np.fft.fftshift(side_freq_theta), np.fft.fftshift(side_freq_r))
        freq_theta_quadrant = freq_theta [r_number // 2:, theta_number // 2:]
        freq_r_quadrant = freq_r  [r_number // 2:, theta_number // 2:]
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.final_diameter<psd_coef["2D"][feature]["p_max"]["D_tie"]:
            p_max = psd_coef["2D"][feature]["p_max"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["p_max"]["b1"]
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_1"]
        else:
            p_max = psd_coef["2D"][feature]["p_max"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["p_max"]["b2"]
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["2D"][feature]["p_diff"]["D_tie"]:
            p_diff = psd_coef["2D"][feature]["p_diff"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["p_diff"]["b1"]
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_1"]
        else:
            p_diff = psd_coef["2D"][feature]["p_diff"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["p_diff"]["b2"]
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["2D"][feature]["nu_fall"]["D_tie"]:
            nu_fall = psd_coef["2D"][feature]["nu_fall"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["nu_fall"]["b1"]
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_1"]
        else:
            nu_fall = psd_coef["2D"][feature]["nu_fall"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["nu_fall"]["b2"]
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_2"]
        if self.crater.final_diameter<psd_coef["2D"][feature]["psd_sigma"]["D_tie"]:
            psd_sigma = psd_coef["2D"][feature]["psd_sigma"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["psd_sigma"]["b1"]
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_1"]
        else:
            psd_sigma = psd_coef["2D"][feature]["psd_sigma"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["psd_sigma"]["b2"]
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_2"]
        p_max = np.random.normal(p_max, p_max_sigma)
        p_diff = np.random.normal(p_diff, p_diff_sigma)
        nu_fall = np.random.normal(nu_fall, nu_fall_sigma)
        psd_sigma = np.random.normal(psd_sigma, psd_sigma_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        freq_r_matrix= np.sqrt(freq_theta_quadrant**2+freq_r_quadrant**2)
        freq_theta_matrix = np.arctan2(freq_r_quadrant,freq_theta_quadrant)
        if feature=="ejecta":
            if self.crater.final_diameter<psd_coef["2D"][feature]["E_rad"]["D_tie"]:
                E_rad = psd_coef["2D"][feature]["E_rad"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["E_rad"]["b1"]
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_1"]
            else:
                E_rad = psd_coef["2D"][feature]["E_rad"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["E_rad"]["b2"]
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_2"]
            E_rad = np.random.normal(E_rad, E_rad_sigma)
            psd_log_quadrant=p_diff/np.sqrt(1-(E_rad*np.sin(freq_theta_matrix))**2)*(np.exp(-np.sqrt(freq_r_matrix/nu_fall))-1)+p_max
        if feature=="wall":
            if self.crater.final_diameter<psd_coef["2D"][feature]["E_circ"]["D_tie"]:
                E_circ = psd_coef["2D"][feature]["E_circ"]["k1"] * self.crater.final_diameter + psd_coef["2D"][feature]["E_circ"]["b1"]
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_1"]
            else:
                E_circ = psd_coef["2D"][feature]["E_circ"]["k2"] * self.crater.final_diameter + psd_coef["2D"][feature]["E_circ"]["b2"]
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_2"]
            E_circ = np.random.normal(E_circ, E_circ_sigma)
            psd_log_quadrant=p_diff/np.sqrt(1-(E_circ*np.cos(freq_theta_matrix))**2)*(np.exp(-np.sqrt(freq_r_matrix/nu_fall))-1)+p_max
        if feature=="floor":
            psd_log_quadrant = p_diff  * (np.exp(-np.sqrt(freq_r_matrix / nu_fall)) - 1) + p_max
        psd_log_quadrant_1st=np.random.normal(psd_log_quadrant,psd_sigma)
        psd_log_quadrant_2nd = np.random.normal(psd_log_quadrant, psd_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        psd_log = np.zeros((r_number, theta_number))
        psd_log[r_number//2:, theta_number // 2:] =  psd_log_quadrant_1st
        psd_log[r_number//2+1:, 1:theta_number // 2] = np.fliplr(psd_log_quadrant_2nd[1:,1:])
        psd_log[ 1:r_number // 2, theta_number // 2+1:] = np.flipud(psd_log_quadrant_2nd[1:, 1:])
        psd_log[1:r_number // 2, 1:theta_number // 2] = np.fliplr(np.flipud(psd_log_quadrant_1st[1:, 1:]))
        #---------------------------------------------------------------------------------------------------------------
        psd_log[1:r_number // 2, theta_number // 2] = np.flip(psd_log[r_number // 2+1:, theta_number // 2])
        psd_log[r_number // 2, 1:theta_number // 2] = np.flip(  psd_log[r_number // 2, theta_number // 2+1:] )
        #---------------------------------------------------------------------------------------------------------------
        psd_log[0,theta_number//2+1:]=psd_log[1,theta_number//2+1:]
        psd_log[r_number // 2 + 1:, 0] = psd_log[r_number // 2 + 1:, 1]
        #---------------------------------------------------------------------------------------------------------------
        psd_log[0, 1:theta_number // 2] = np.flip(psd_log[0,theta_number//2+1:])
        psd_log[1:r_number // 2:, 0] = np.flip(psd_log[r_number // 2 + 1:,0] )
        # --------------------------------------------------------------------------------------------------------------
        psd_log[0, 0]=psd_log[1, 1]
        psd_log[0, theta_number//2] =psd_log[1, theta_number//2]
        psd_log[r_number//2, 0] = psd_log[r_number//2, 1]
        # ------------------------------------------------------------------------------------------------------------------
        psd = 10 ** psd_log
        phase = np.random.uniform(-math.pi, math.pi, size=(r_number, theta_number))
        if feature=="ejecta" or feature=="wall" : psd[:,theta_number//2]=np.zeros((r_number))
        # ------------------------------------------------------------------------------------------------------------------
        freq_theta_delta=(freq_theta.max()-freq_theta.min())/theta_number
        freq_r_delta = (freq_r.max() - freq_r.min()) / r_number
        freq_theta_num=int(max_effec_freq/freq_theta_delta)
        freq_r_num = int(max_effec_freq / freq_r_delta)
        freq_theta = freq_theta[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        freq_r = freq_r[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        psd = psd[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        phase = phase[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        # ------------------------------------------------------------------------------------------------------------------
        psd=np.sqrt(psd*area)
        return psd,phase,freq_theta,freq_r,grid_theta,grid_r  



from numpy.random import Generator
from typing import Any
from cratermaker.components.target import Target
from cratermaker.components.projectile import Projectile
from cratermaker.utils.custom_types import FloatLike
from warnings import warn

@Projectile.register("asteroids")
class AsteroidProjectiles(Projectile):
    def __init__(self, 
                 target : Target | str | None = None,
                 density : FloatLike | None = None,
                 rng: Generator | None = None,
                 rng_seed : int | None = None,
                 rng_state : dict | None = None, 
                 **kwargs: Any):
        """
        An operations class for computing the projectile properties of an asteroid source population.

        Parameters
        ----------
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        density : float
            The density of the projectile in kg/m^3. Default is 2250.0 kg/m^3.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """

        # This model always samples velocities, angles, and directions, so override any values that may have been passed.
        if density is None:
            density = 2250.0
        super().__init__(target=target, density=density, rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        self.mean_velocity = self._set_mean_velocity()


    def _set_mean_velocity(self):
        """
        Sets the mean velocity of the projectile in m/s based on the target body.
        
        Returns
        -------
        float 
        """
        known_targets = ['Mercury', 'Venus', 'Earth', 'Moon', 'Mars', 'Ceres', 'Vesta', 'MBA']
        target_velocities = [41100.0, 29100.0, 24600.0, 22100.0, 10700.0, 5300.0, 5300.0, 5300.0]
        catalogue = dict(zip(known_targets, target_velocities))
        if self.target_name in known_targets:
            pmv = float(catalogue[self.target_name])
        else:
            warn(f"Target {self.target_name} not found in known targets. Known targets include {known_targets}. Defaulting to the Moon.")
            pmv = float(catalogue["Moon"])
        return pmv



"""
Plot isochrons for the Moon and Mars for 1 Ma, 1 Ga, and 4 Ga using the Neukum Production Function
==================================================================================================

.. rubric:: By Austin Blevins and David Minton

In this example, we will be using the "neukum" prodcution model in Cratermaker to plot isochrons for three different age surfaces. We will also format the plots so with similar axes as Figure 2 of Neukum, Ivanov, and Hartmann (2001) [#]_. 

.. [#] Neukum, G., Ivanov, B.A., Hartmann, W.K., 2001. Cratering Records in the Inner Solar System in Relation to the Lunar Reference System. Space Science Reviews 96, 55–86. https://doi.org/10.1023/A:1011989004263
"""

from cratermaker import Production
import matplotlib.pyplot as plt
from matplotlib import ticker 
import numpy as np

fig, axs = plt.subplots(1, 2, figsize=(8, 7))
axes = dict(zip(['Moon','Mars'],axs))
tvals = [0.01, 1.0, 4.0]
x_min = 1e-3
x_max = 1e4
y_min = 1e-12
y_max = 1e6
nD = 1000
Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
for version, ax in axes.items():
    production = Production.maker("neukum", version=version)
    ax.title.set_text(version)
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_ylabel('$\\mathregular{N_{>D} (km^{-2})}$')
    ax.set_xlabel('Diameter (km)')
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
    ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
    ax.grid(True,which="minor",ls="-",lw=0.5,zorder=5)
    ax.grid(True,which="major",ls="-",lw=1,zorder=10)
    inrange = (Dvals > production.sfd_range[0]*1e-3) & (Dvals < production.sfd_range[1]*1e-3)
    lo = Dvals < production.sfd_range[0]*1e-3
    hi = Dvals > production.sfd_range[1]*1e-3
    for t in tvals:
        Nvals = production.function(diameter=Dvals*1e3,age=t*1e3)
        Nvals = Nvals * 1e6 
        ax.plot(Dvals[inrange], Nvals[inrange], '-', color='black', linewidth=1.0, zorder=50)
        ax.plot(Dvals[lo], Nvals[lo], '-.', color='orange', linewidth=2.0, zorder=50)
        ax.plot(Dvals[hi], Nvals[hi], '-.', color='orange', linewidth=2.0, zorder=50)
        labeli = int(0.25*nD)
        ax.text(Dvals[labeli],3*Nvals[labeli],f"{t:.2f} ", ha="left", va="top",rotation=-72)

plt.tight_layout()
plt.show()


from __future__ import annotations
from typing import Any
import numpy as np
import math
from numpy.random import Generator
from cratermaker.utils.general_utils import parameter, format_large_units, _set_properties
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker.utils.custom_types import FloatLike
from cratermaker.utils import montecarlo_utils as mc
from cratermaker.components.target import Target

class Projectile(ComponentBase):
    _registry: dict[str, Projectile] = {}
    def __init__(self, 
                 target : Target | str | None = None,
                 mean_velocity : FloatLike | None = None,
                 density : FloatLike | None = None,
                 sample : bool | None = None,
                 angle: FloatLike | None = None,
                 velocity : FloatLike | None = None,
                 direction : FloatLike | None = None,
                 rng: Generator | None = None,
                 rng_seed : int | None = None,
                 rng_state : dict | None = None, 
                 **kwargs):
        """
        This is the abstract base class for all projectile models. It defines the interface for generating projectile velocities, angles, and densities for a given target body.

        Parameters
        ----------
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        mean_velocity : float
            The mean velocity of the projectile in m/s. 
        density : float
            The density of the projectile in kg/m^3. 
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to False, impact velocities will be set to the mean velocity, impact angles will be set to 90 degrees (vertical impact), and directions will be 0.
        angle : float
            The impact angle in degrees. 
        velocity : float | None
            The impact velocity in m/s. If None, the velocity will be sampled from a distribution.
        direction : float | None
            The impact direction in degrees. If None, the direction will be sampled from a distribution.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_target", target)
        object.__setattr__(self, "_sample", sample)
        object.__setattr__(self, "_mean_velocity", mean_velocity)
        object.__setattr__(self, "_density", density)
        object.__setattr__(self, "_velocity", velocity)
        object.__setattr__(self, "_direction", direction)
        object.__setattr__(self, "_angle", angle)
        if self.sample is None:
            self.sample = True
        self.target = Target.maker(target, **kwargs)

    def __repr__(self) -> str:
        base = super().__repr__()
        mean_velocity = format_large_units(self.mean_velocity, quantity="velocity")
        return (
            f"{base}\n"
            f"Target: {self.target.name}\n"
            f"Density: {self.density:.1f} kg/m³\n"
            f"Mean Velocity: {mean_velocity}\n"
            f"Sample from distributions: {self.sample}"
        )


    @classmethod
    def maker(cls,
             projectile: Projectile | str | None = None, 
             target : Target | str | None = None,
             mean_velocity : FloatLike | None = None,
             density : FloatLike | None = None,
             sample : bool = True,
             angle: FloatLike | None = None,
             velocity : FloatLike | None = None,
             direction : FloatLike | None = None,
             rng: Generator | None = None,
             rng_seed : int | None = None,
             rng_state : dict | None = None, 
             **kwargs: Any) -> Projectile:
        """
        Initialize an projectile model based on the provided name or class.
        
        Parameters
        ----------
        projectile : Projectile or str
            The projectile model to initialize. Can be a class or a string representing the model name.
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        mean_velocity : float
            The mean velocity of the projectile in m/s. 
        density : float
            The density of the projectile in kg/m^3. 
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to False, impact velocities will be set to the mean velocity, impact angles will be set to 90 degrees (vertical impact), and directions will be 0.
        angle : float
            The impact angle in degrees. 
        velocity : float | None
            The impact velocity in m/s. If None, the velocity will be sampled from a distribution.
        direction : float | None
            The impact direction in degrees. If None, the direction will be sampled from a distribution.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        
        Returns
        -------
        Projectile
            The initialized projectile model.
        
        Raises
        ------
        KeyError
            If the specified projectile model is not found in the registry.
        TypeError
            If the specified projectile model is not a string or a subclass of Projectile.
        """
        target = Target.maker(target, **kwargs)
        if projectile is None:
            target_name = target.name.capitalize()
            if target_name in ['Mercury', 'Venus', 'Earth', 'Moon', 'Mars', 'Ceres', 'Vesta']:
                projectile = "asteroids"
            else:
                projectile = "comets"

        projectile = super().maker(component=projectile,
                            target=target,
                            mean_velocity=mean_velocity,
                            density=density,
                            sample=sample,
                            angle=angle,
                            velocity=velocity,
                            direction=direction,
                            rng=rng,
                            rng_seed=rng_seed,
                            rng_state=rng_state, 
                            **kwargs) 
        projectile.new_projectile(**kwargs)
        return projectile

        
    def new_projectile(self, **kwargs: Any) -> dict:
        """
        Updates the values of the velocities and angles and returns them as a dictionary of projectile properties that can be passed as arguments to the Crater class.
        
        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments to be passed to internal functions.
        
        Returns
        -------
        dict
            A dictionary containing the projectile properties.
        """

        if self.sample:
            self._velocity = float(mc.get_random_velocity(self.mean_velocity, rng=self.rng)[0])
        elif self.velocity is None:
            self._velocity = float(self.mean_velocity)

        if self.sample:
            self._angle = float(mc.get_random_impact_angle(rng=self.rng)[0])
        elif self.angle is None:
            self._angle = float(90.0)

        if self.sample:
            self._direction = float(mc.get_random_impact_direction(rng=self.rng)[0])
        elif self._direction is None:
            self._direction = float(0.0)

        return {
            "projectile_velocity": self.velocity,
            "projectile_angle": self.angle,
            "projectile_density": self.density,
            "projectile_direction": self.direction, 
        }

    @parameter
    def target_name(self):
        """
        The name of the target body.
        
        Returns
        -------
        str
        """ 
        return self._target.name
    
    @property
    def target(self):
        """
        The target object for the projectile model.
        
        Returns
        -------
        Target
        """
        return self._target
    
    @target.setter
    def target(self, value):
        self._target = Target.maker(value)

    @parameter
    def sample(self):
        """
        Flag that determines whether to sample velocities, angles, and directions from distributions. If set to False, impact velocities will be set to the mean velocity, impact angles will be set to 90 degrees (vertical impact), and directions will be 0.
        
        Returns
        -------
        bool
        """
        return self._sample
    
    @sample.setter
    def sample(self, value):
        if not isinstance(value, bool):
            raise TypeError("sample must be a boolean value")
        self._sample = value
        return
    
    @parameter
    def mean_velocity(self):
        """
        The mean velocity of the projectile in m/s.
        
        Returns
        -------
        float 
        """
        return self._mean_velocity
    
    @mean_velocity.setter
    def mean_velocity(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if isinstance(value, FloatLike):
            if value < 0:
                raise ValueError("mean_velocity must be a positive number")
            self._mean_velocity = float(value)
        else: 
            raise TypeError("mean_velocity must be a numeric value") 
        return

    @property 
    def angle(self):
        """
        The impact angle in degrees.
        
        Returns
        -------
        float 
        """
        return self._angle
    
    @angle.setter
    def angle(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if not isinstance(value, FloatLike):
            raise TypeError("angle must be a numeric value")
        if value < 0:
            raise ValueError("angle must be a positive number")
        self._angle = float(value)

    @property
    def direction(self):
        """
        The impact direction in degrees.
        
        Returns
        -------
        float 
        """
        return self._direction
    
    @direction.setter
    def direction(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if not isinstance(value, FloatLike):
            raise TypeError("direction must be a numeric value")
        if value < 0:
            raise ValueError("direction must be a positive number")
        self._direction = float(value)

    @property
    def velocity(self):
        """
        The impact velocity in m/s.
        
        Returns
        -------
        float 
        """
        return self._velocity
    
    @velocity.setter
    def velocity(self, value):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("velocity must be a numeric value")
            if value < 0:
                raise ValueError("velocity must be a positive number")
            self._velocity = float(value)
        else:
            self._velocity = None
        return

    @parameter
    def density(self):
        """
        The density of the projectile in kg/m^3.
        
        Returns
        -------
        float 
        """
        return self._density
    
    @density.setter
    def density(self, value):
        """
        Sets the density of the projectile in kg/m^3. 
        
        Parameters
        ----------
        value : float | None
            The density in kg/m^3 or None to use a default value.
        
        Raises
        ------
        ValueError
            If the provided value is negative or None.
        TypeError
            If the provided value is not a numeric value or None.
        """
        if not isinstance(value, FloatLike):
            raise TypeError("density must be a numeric value")
        if value < 0:
            raise ValueError("density must be a positive number")
        self._density = float(value)

    @property
    def vertical_velocity(self):
        """
        The vertical component of the impact velocity in m/s.
        
        Returns
        -------
        float 
        """
        return self.velocity * math.sin(math.radians(self.angle))
    
    @property
    def population(self):
        """
        The name of the population of the projectile model.
        
        Returns
        -------
        int
        """
        return self._component_name

import_components(__name__, __path__, ignore_private=True)




import numpy as np
import xarray as xr
from tqdm import tqdm
from pathlib import Path
from numpy.typing import ArrayLike
from cratermaker.utils.custom_types import FloatLike
from cratermaker.components.surface import Surface
from cratermaker.constants import _CIRCLE_FILE_NAME, _EXPORT_DIR, _DATA_DIR, _COMBINED_DATA_FILE_NAME, _VTK_FILE_EXTENSION

def to_vtk(surface : Surface, 
                *args, **kwargs
                ) -> None:
    """
    Export the surface mesh to a VTK file and stores it in the default export directory.
    """
    from vtk import vtkUnstructuredGrid, vtkPoints, VTK_POLYGON, vtkWarpScalar, vtkXMLPolyDataWriter
    from vtkmodules.util.numpy_support import numpy_to_vtk
    from vtkmodules.vtkFiltersCore import vtkPolyDataNormals
    from vtkmodules.vtkFiltersGeometry import vtkGeometryFilter

    if not isinstance(surface, Surface):
        raise TypeError("The surface argument must be an instance of the Surface class.") 
    # Create the output directory if it doesn't exist 
    out_dir = surface.simdir / _EXPORT_DIR
    out_dir.mkdir(parents=True, exist_ok=True)

    data_dir = surface.simdir / _DATA_DIR
    data_file_list = list(data_dir.glob("*.nc"))
    if surface.grid_file in data_file_list:
        data_file_list.remove(surface.grid_file)

    # Delete old export files if they exist
    old_export_files = list(out_dir.glob(f"*.{_VTK_FILE_EXTENSION}"))
    if len(old_export_files) > 0:
        for old_file in old_export_files:
            old_file.unlink()
    
    # Convert uxarray grid arrays to regular numpy arrays for vtk processing 
    n_node = surface.uxgrid.n_node
    n_face = surface.uxgrid.n_face
    node_x = surface.uxgrid.node_x.values * surface.target.radius
    node_y = surface.uxgrid.node_y.values * surface.target.radius
    node_z = surface.uxgrid.node_z.values * surface.target.radius
    n_nodes_per_face = surface.uxgrid.n_nodes_per_face.values
    face_node_connectivity = surface.uxgrid.face_node_connectivity.values
    
    vtk_data = vtkUnstructuredGrid()
    nodes = vtkPoints()
    for i in range(n_node):
        nodes.InsertNextPoint(node_x[i], node_y[i], node_z[i])
    vtk_data.SetPoints(nodes)
    vtk_data.Allocate(n_face)
    for i,n in enumerate(n_nodes_per_face):
        point_ids=face_node_connectivity[i][0:n]
        vtk_data.InsertNextCell(VTK_POLYGON, n, point_ids) 
    
    warp = vtkWarpScalar()
    warp.SetInputArrayToProcess(0, 0, 0,
                        vtkUnstructuredGrid.FIELD_ASSOCIATION_POINTS,
                        "node_elevation")                
        
    writer = vtkXMLPolyDataWriter()
    writer.SetDataModeToBinary()
    writer.SetCompressorTypeToZLib()
    print("Exporting VTK files...")
    
    with xr.open_mfdataset(data_file_list) as ds:
        # Warp the surface based on node_elevation data
        for i in tqdm(range(len(ds.time))):
            
            ids = ds.isel(time=i).load()
            current_grid = vtkUnstructuredGrid()
            current_grid.DeepCopy(vtk_data) 
                
            for v in ds.variables:
                array = numpy_to_vtk(ids[v].values, deep=True)
                array.SetName(v)
                n = ids[v].size
                if 'n_face' in ids[v].dims:
                    current_grid.GetCellData().AddArray(array)
                elif 'n_node' in ids[v].dims:
                    current_grid.GetPointData().AddArray(array)
                    if v == 'node_elevation':
                        current_grid.GetPointData().SetActiveScalars(v) 
                elif n == 1:
                    current_grid.GetFieldData().AddArray(array)
                    
            geomFilter = vtkGeometryFilter()
            geomFilter.SetInputData(current_grid)
            geomFilter.Update()    
            polyData = geomFilter.GetOutput()    

            normalsFilter = vtkPolyDataNormals()
            normalsFilter.SetInputData(polyData)
            normalsFilter.ComputeCellNormalsOn()
            normalsFilter.ConsistencyOn()           # Tries to make normals consistent across shared edges
            normalsFilter.AutoOrientNormalsOn()     # Attempt to orient normals consistently outward/inward
            normalsFilter.SplittingOff()   
            normalsFilter.Update()
            polyDataWithNormals = normalsFilter.GetOutput()        
    
            warp.SetInputData(polyDataWithNormals)
            warp.Update()
            warped_output = warp.GetOutput()
            output_filename = out_dir / _COMBINED_DATA_FILE_NAME.replace(".nc", f"{i:06d}.{_VTK_FILE_EXTENSION}")
            writer.SetFileName(output_filename)
            writer.SetInputData(warped_output) 
            writer.Write()               
    
    return


def make_circle_file(surface,
                        diameters: FloatLike | ArrayLike, 
                        longitudes: FloatLike | ArrayLike,
                        latitudes: FloatLike | ArrayLike,
                        output_filename: str | Path | None = None,
                        *args, **kwargs
                    ) -> None:
    """
    Plot circles of diameter D centered at the given location.

    Parameters
    ----------
    diameters : FloatLike or ArrayLike
        Diameters of the circles in m.
    longitudes : FloatLike or ArrayLike of Floats
        Longitudes of the circle centers in degrees.
    latitudes : FloatLike or ArrayLike of Floats
        Latitudes of the circle centers in degrees.
    out_filename : str or Path, optional 
        Name of the output file. If not provided, the default is "circle.vtp" 
    """ 
    from vtk import vtkPoints, vtkPolyLine, vtkCellArray, vtkPolyData, vtkXMLPolyDataWriter

    if output_filename is None:
        output_filename = surface.simdir / _EXPORT_DIR / _CIRCLE_FILE_NAME
    else:
        output_filename = surface.simdir / _EXPORT_DIR / output_filename
    
    diameters = np.atleast_1d(diameters)
    longitudes = np.atleast_1d(longitudes)
    latitudes = np.atleast_1d(latitudes)
    # Check for length consistency
    if len(diameters) != len(longitudes) or len(diameters) != len(latitudes):
            raise ValueError("The diameters, latitudes, and longitudes, arguments must have the same length")
        
    # Validate non-negative values
    if np.any(diameters < 0):
        raise ValueError("All values in 'diameters' must be non-negative")
    
    sphere_radius = surface.target.radius 
    def create_circle(lon, lat, circle_radius, num_points=360):
        """
        Create a circle on the sphere's surface with a given radius and center.
        
        Parameters
        ----------
        lon : float
            Longitude of the circle's center in degrees.
        lat : float
            Latitude of the circle's center in degrees.
        circle_radius : float
            Radius of the circle in meters.
        num_points : int, optional
            Number of points to use to approximate the circle. The default is 360. 
        """
        # Create an array of angle steps for the circle
        radians = np.linspace(0, 2 * np.pi, num_points)
        # Convert latitude and longitude to radians
        lat_rad = np.deg2rad(lat)
        lon_rad = np.deg2rad(lon)

        # Calculate the Cartesian coordinates for the circle's center
        center_x = sphere_radius * np.cos(lat_rad) * np.cos(lon_rad)
        center_y = sphere_radius * np.cos(lat_rad) * np.sin(lon_rad)
        center_z = sphere_radius * np.sin(lat_rad)

        # Calculate the vectors for the local east and north directions on the sphere's surface
        east = np.array([-np.sin(lon_rad), np.cos(lon_rad), 0])
        north = np.array([-np.cos(lon_rad)*np.sin(lat_rad), -np.sin(lon_rad)*np.sin(lat_rad), np.cos(lat_rad)])
        
        # Initialize arrays to hold the circle points
        x = np.zeros_like(radians)
        y = np.zeros_like(radians)
        z = np.zeros_like(radians)

        # Calculate the points around the circle
        for i in range(num_points):
            x[i] = center_x + circle_radius * np.cos(radians[i]) * east[0] + circle_radius * np.sin(radians[i]) * north[0]
            y[i] = center_y + circle_radius * np.cos(radians[i]) * east[1] + circle_radius * np.sin(radians[i]) * north[1]
            z[i] = center_z + circle_radius * np.cos(radians[i]) * east[2] + circle_radius * np.sin(radians[i]) * north[2]

        return x, y, z 

    points = vtkPoints()
    lines = vtkCellArray()
    point_id = 0  # Keep track of the point ID across all circles
    
    for lon, lat, diameter in zip(longitudes, latitudes, diameters):
        circle_radius = diameter / 2
        x, y, z = create_circle(lon, lat, circle_radius)
        
        for i in range(len(x)):
            points.InsertNextPoint(x[i], y[i], z[i])
        
        polyline = vtkPolyLine()
        polyline.GetPointIds().SetNumberOfIds(len(x))
        for i in range(len(x)):
            polyline.GetPointIds().SetId(i, point_id)
            point_id += 1
        
        lines.InsertNextCell(polyline)
    
    # Create a polydata object and add points and lines to it
    polydata = vtkPolyData()
    polydata.SetPoints(points)
    polydata.SetLines(lines)
    
    # Write the polydata to a VTK file
    writer = vtkXMLPolyDataWriter()
    writer.SetFileName(output_filename)
    writer.SetInputData(polydata)
    
    # Optional: set the data mode to binary to save disk space
    writer.SetDataModeToBinary()
    writer.Write()
    
    return    
